@@ -2,18 +2,18 @@
 
 """ownCloud and Seafile conflict resolver
 
-ownCloud often generates bunch of conflict files. Seafile is far better, but it
-happens. Some of conflict files are completely senseless (they are equivalent to
-non conflict files laying beside them), some of them really differs from their
-neighborhoods. The aim of this script is to resolve conflict files of both
-types.
-"""
+ownCloud often generates bunch of conflict files. Seafile is far better,
+but it happens. Some of conflict files are completely senseless (they are
+equivalent to non conflict files laying beside them), some of them really
+differs from their neighborhoods. The aim of this script is to resolve
+conflict files of both types. """
 
-__version__ = "0.0.4"
+__version__ = "0.0.5"
 
 # pylint: disable=C0103, C0111, W0105, W0141, W0142
 
 import argparse
+import errno
 import filecmp
 import logging as logger
 import os
@@ -23,7 +23,6 @@ import shutil
 import string
 import subprocess
 import tarfile
-from collections import OrderedDict
 from pathlib import Path
 try:
     from enum import IntEnum
@@ -44,6 +43,39 @@ class Keep(IntEnum):
     both = 3
 
 
+programs = {
+    'merge': {
+        'kdiff3': {
+            'arguments': '-m $conflict $normal -o $normal',
+            'outputs': False,
+            'wait': True,
+        },
+    },
+    'compare': {
+        'diff': {
+            'arguments': '--side-by-side $conflict $normal',
+            'outputs': True,
+            'wait': True,
+        },
+        'vimdiff': {
+            'arguments': '$conflict -c \':vertical diffsplit $normal\'',
+            'outputs': True,
+            'wait': True,
+        },
+        'meld': {
+            'arguments': '$conflict $normal',
+            'outputs': False,
+            'wait': False,
+        },
+        'kdiff3': {
+            'arguments': '$conflict $normal',
+            'outputs': False,
+            'wait': False,
+        },
+    },
+}
+
+
 class FileManipulatorMeta(type):
     def __str__(cls):
         return cls.__name__.replace(cls.__base__.__name__, '').lower()
@@ -138,22 +170,36 @@ class StopResolving(ResolverException):
 
 class ConflictResolver:
     class Program:
-        default_name = "kdiff3"
-
-        def __init__(self, name=None, arguments=None):
-            self.name = name if name else self.default_name
+        def __init__(self, name=None, *, arguments=None, outputs=False,
+                     wait=True):
+            self.name = name
             self.arguments = arguments
+            self.outputs = outputs
+            self.wait = wait
             self.process = None
 
-        def __bool__(self):
+        @property
+        def exists(self):
             return bool(shutil.which(self.name))
 
         def execute(self, conflict_file_path):
             normal_file_path = ConflictResolver.normal_file(conflict_file_path)
+            print("1: \"{}\" vs 2. \"{}\"".format(
+                conflict_file_path,  normal_file_path))
             program_args = [arg.safe_substitute(normal=normal_file_path,
                                                 conflict=conflict_file_path)
                             for arg in self.arguments]
-            self.process = subprocess.Popen([self.name] + program_args)
+            self.process = subprocess.Popen(
+                [self.name] + program_args,
+                stdout=subprocess.DEVNULL if not self.outputs else None,
+                stderr=subprocess.DEVNULL)
+            return_code = 0
+            if self.wait:
+                return_code = self.process.wait()
+                self.process = None
+            if self.outputs:
+                print()
+            return return_code
 
         def terminate(self):
             if self.process:
@@ -163,13 +209,14 @@ class ConflictResolver:
             if self.process:
                 return self.process.wait()
 
+    resolving_type = 'auto'
     re_conflict_file_path_part = {
         'owncloud': r"_conflict-\d{8,8}-\d{6,6}",
         'seafile': r" \(.*@.* \d{4}-[01]\d-[0-3]\d-[0-2]\d-[0-5]\d-[0-5]\d\)",
     }
 
-    def __init__(self, root=Path('.'), *, deleter=RemoveDeleter, renamer=Renamer,
-                 program=Program(), backup=True):
+    def __init__(self, root=Path('.'), *, deleter=RemoveDeleter,
+                 renamer=Renamer, program=Program(), backup=True):
         self.root = root
         self.deleter = deleter()
         self.renamer = renamer()
@@ -207,8 +254,8 @@ class ConflictResolver:
                     self.deleter.delete(conflict_file_path)
             else:
                 for symlink in symlinks:
-                    logger.info('File {} symlink is symlink. Removing symlinks '
-                                'is not implemented'.format(symlink))
+                    logger.info('File {} symlink is symlink. Removing '
+                                'symlinks is not implemented'.format(symlink))
                 result = True
         else:
             logger.debug(
@@ -240,10 +287,13 @@ class ConflictResolver:
             try:
                 os.chdir(str(self.root))
                 for conflict_file in self.conflict_files():
-                    backup_file.add(str(conflict_file.relative_to(self.root)))
+                    def relative_to_root(f):
+                        return str(f.relative_to(self.root))
+
+                    backup_file.add(relative_to_root(conflict_file))
                     normal_file = self.normal_file(conflict_file)
                     if normal_file.exists():
-                        backup_file.add(str(normal_file.relative_to(self.root)))
+                        backup_file.add(relative_to_root(normal_file))
             finally:
                 os.chdir(save_cwd)
 
@@ -256,12 +306,6 @@ class ConflictResolver:
 
         if self.backup:
             self.create_backup()
-        if not self.program:
-            if self.program is not None:
-                logger.error('Program "{}" not found, please install it or '
-                             "specify different program with option "
-                             '"--program"'.format(self.program.name))
-            exit()
         try:
             for conflict_file_path in self.conflict_files():
                 try:
@@ -274,6 +318,8 @@ class ConflictResolver:
 
 
 class CompareConflictResolver(ConflictResolver):
+    resolving_type = 'compare'
+
     def resolve_manually(self, conflict_file_path, *, keep=Keep.both):
         result = True
         normal_file_path = self.normal_file(conflict_file_path)
@@ -301,14 +347,15 @@ class CompareConflictResolver(ConflictResolver):
                     def print_help():
                         print()
                         print("0 - don't save any file, remove both files\n"
-                              "1 - remove normal file, save and rename conflict file\n"
+                              "1 - remove normal file, save and rename "
+                              "conflict file\n"
                               "2 - save normal file, remove conflict file\n"
                               "r - reset all actions, leave all files as are\n"
                               "s - stop processing and apply all actions\n")
 
                     answer = input(
-                        "What file do you want to save (1 - conflict, 2 - normal, "
-                        "3 - both, ? - show all options) [3]: ")
+                        "What file do you want to save (1 - conflict, "
+                        "2 - normal, 3 - both, ? - show all options) [3]: ")
                     if answer == '':
                         keep = Keep.both
                     elif answer == '?':
@@ -329,17 +376,20 @@ class CompareConflictResolver(ConflictResolver):
 
 
 class MergeConflictResolver(ConflictResolver):
+    resolving_type = 'merge'
+
     def resolve(self, conflict_file_path):
         result = super().resolve(conflict_file_path)
-        if not result and self.program:
-            self.program.execute(conflict_file_path)
-            result = self.program.wait() == 0
+        if not result:
+            result = self.program.execute(conflict_file_path) == 0
             if result:
                 self.deleter.delete(conflict_file_path)
         return result
 
 
 class ListConflictResolver(ConflictResolver):
+    resolving_type = 'none'
+
     def __init__(self, root=Path('.'), *, list_normal=False, **kwargs):
         super().__init__(root, **kwargs)
         self.list_normal = list_normal
@@ -358,6 +408,18 @@ class ListAllConflictResolver(ListConflictResolver):
         super().__init__(root, list_normal=True, **kwargs)
 
 
+def check_program(program):
+    if not program.exists:
+        logger.error(
+            "Program \"{}\" not found. You have several options:\n"
+            "1. Install it.\n"
+            "2. Specify different program with option \"--program-name\".\n"
+            "3. Ignore \"hard\" files (what cannot be automatically "
+            "resolved) by using option \"--non-interactive\"."
+            "".format(program.name))
+        exit(errno.ENOENT)
+
+
 def to_(s, delimiter='_'):
     return re.sub(r'(?!^)([A-Z]+)', delimiter + r'\1', s).lower()
 
@@ -440,37 +502,41 @@ def parse_arguments():
                                          help="list conflict files and "
                                               "their corresponding normal "
                                               "files")
-    parser.add_argument('-p', '--program', default='',
-                        help="program to compare files [default: kdiff3]")
-    parser.add_argument('-a', '--arguments', default='',
+    parser.add_argument('-p', '--program-name', default='kdiff3',
+                        help="program to compare files; diff, vimdiff, meld, "
+                             "and kdiff3 work out of box, use "
+                             "--program-arguments for other [default: kdiff3]")
+    parser.add_argument('-a', '--program-arguments',
                         help="arguments that should be passed to conflict "
                              "resolver program; you can use predefined "
-                             "variables (use brackets) to substitute with file "
-                             "name: ($conflict — file with conflict appendix, "
-                             "$normal — file without conflict appendix)")
+                             "variables (use brackets) to substitute with "
+                             "file name: ($conflict — file with conflict "
+                             "appendix, $normal — file without conflict "
+                             "appendix)")
     backup_group = parser.add_mutually_exclusive_group()
     backup_group.add_argument('-b', '--backup', action="store_true",
                               default=True, help="create backup [default]")
     backup_group.add_argument('-B', '--no-backup', action="store_false",
-                              dest="backup", help="don't create backup")
+                              dest="backup",
+                              help="don't create backup (not recommended)")
     parser.set_defaults(conflict_resolver=CompareConflictResolver,
                         deleter=RemoveDeleter)
     if not can_delete_to_trash:
         parser.epilog += " To enable deleting to trash, please install python "
-        "module send2trash. You can do it with pip:\n"
-        "  pip3 install send2trash"
+        "package send2trash."
     result = parser.parse_args()
     result.dirs = [Path(d) for d in result.dirs]
     if result.deleter is not LogDeleter:
         result.renamer = Renamer
     else:
         result.renamer = LogRenamer
-    if not result.arguments:
-        if result.conflict_resolver is MergeConflictResolver:
-            result.arguments = "-m $conflict $normal -o $normal"
-        else:
-            result.arguments = "$conflict $normal"
-    result.arguments = [string.Template(arg) for arg in shlex.split(result.arguments)]
+    result.program = (
+        programs[result.conflict_resolver.resolving_type][result.program_name])
+    result.program['name'] = result.program_name
+    slr_arguments = (result.program_arguments if result.program_arguments else
+                     result.program['arguments'])
+    result.program['arguments'] = [
+        string.Template(arg) for arg in shlex.split(slr_arguments)]
     return result
 
 
@@ -478,7 +544,8 @@ def run(args):
     for dir_path in args.dirs:
         log_level = max(logger.INFO - 10*args.verbose_count, logger.NOTSET)
         logger.basicConfig(level=log_level, format='%(message)s')
-        program = ConflictResolver.Program(args.program, args.arguments)
+        program = ConflictResolver.Program(**args.program)
+        check_program(program)
         logger.debug("Using {}".format(
             to_(args.conflict_resolver.__name__, delimiter=' ')))
         resolver = args.conflict_resolver(
